//
// Copyright (C)  2005-2008 Greg Landrum and Rational Discovery LLC
//
//   @@ All Rights Reserved @@
//  This file is part of the RDKit.
//  The contents are covered by the terms of the BSD license
//  which is included in the file license.txt, found at the root
//  of the RDKit source tree.
//
#include "BFGSOpt.h"

namespace BFGSOpt {
int REALLY_A_HEADER_ONLY_LIBRARY = 0;
}
